@@ -1,5 +1,5 @@
 # --- BEGIN COPYRIGHT BLOCK ---
-# Copyright (C) 2020 Red Hat, Inc.
+# Copyright (C) 2022 Red Hat, Inc.
 # Copyright (C) 2019 William Brown <william@blackhats.net.au>
 # All rights reserved.
 #
@@ -30,19 +30,23 @@ from lib389.configurations.sample import (
     create_base_cn,
     create_base_c,
 )
-from lib389.instance.options import General2Base, Slapd2Base, Backend2Base
+from lib389.instance.options import General2Base, Slapd2Base, Backend2Base, Options2
 from lib389.paths import Paths
 from lib389.saslmap import SaslMappings
 from lib389.instance.remove import remove_ds_instance
 from lib389.index import Indexes
+from lib389.replica import Replicas, BootstrapReplicationManager, Changelog
 from lib389.utils import (
     assert_c,
     is_a_dn,
     ensure_str,
     ensure_list_str,
+    get_default_db_lib,
     normalizeDN,
     socket_check_open,
+    selinux_label_file,
     selinux_label_port,
+    resolve_selinux_path,
     selinux_restorecon,
     selinux_present)
 
@@ -195,6 +199,25 @@ class SetupDs(object):
                     if req_idx:
                         be[BACKEND_REQ_INDEX] = "on"
 
+                    # Replication settings
+                    be[BACKEND_REPL_ENABLED] = False
+                    if config.get(section, BACKEND_REPL_ENABLED, fallback=False):
+                        be[BACKEND_REPL_ENABLED] = True
+                        role = config.get(section, BACKEND_REPL_ROLE, fallback="supplier")
+                        be[BACKEND_REPL_ROLE] = role
+                        rid = config.get(section, BACKEND_REPL_ID, fallback="1")
+                        be[BACKEND_REPL_ID] = rid
+                        binddn = config.get(section, BACKEND_REPL_BINDDN, fallback=None)
+                        be[BACKEND_REPL_BINDDN] = binddn
+                        bindpw = config.get(section, BACKEND_REPL_BINDPW, fallback=None)
+                        be[BACKEND_REPL_BINDPW] = bindpw
+                        bindgrp = config.get(section, BACKEND_REPL_BINDGROUP, fallback=None)
+                        be[BACKEND_REPL_BINDGROUP] = bindgrp
+                        cl_max_entries = config.get(section, BACKEND_REPL_CL_MAX_ENTRIES, fallback="-1")
+                        be[BACKEND_REPL_CL_MAX_ENTRIES] = cl_max_entries
+                        cl_max_age = config.get(section, BACKEND_REPL_CL_MAX_AGE, fallback="7d")
+                        be[BACKEND_REPL_CL_MAX_AGE] = cl_max_age
+
                     # Add this backend to the list
                     backends.append(be)
 
@@ -253,8 +276,8 @@ class SetupDs(object):
         general = {'config_version': 2,
                    'full_machine_name': socket.getfqdn(),
                    'strict_host_checking': False,
-                   'selinux': True,
-                   'systemd': ds_paths.with_systemd,
+                   'selinux': Options2.default_values['selinux'],
+                   'systemd': Options2.get_systemd_default(),
                    'defaults': '999999999', 'start': True}
 
         slapd = {'self_sign_cert_valid_months': 24,
@@ -263,10 +286,10 @@ class SetupDs(object):
                  'initconfig_dir': ds_paths.initconfig_dir,
                  'self_sign_cert': True,
                  'root_password': '',
-                 'port': 389,
+                 'port': Options2.default_values['port'],
                  'instance_name': 'localhost',
                  'user': ds_paths.user,
-                 'secure_port': 636,
+                 'secure_port': Options2.default_values['secure_port'],
                  'prefix': ds_paths.prefix,
                  'bin_dir': ds_paths.bin_dir,
                  'sbin_dir': ds_paths.sbin_dir,
@@ -283,6 +306,7 @@ class SetupDs(object):
                  'backup_dir': ds_paths.backup_dir,
                  'db_dir': ds_paths.db_dir,
                  'db_home_dir': ds_paths.db_home_dir,
+                 'db_lib': get_default_db_lib(),
                  'ldif_dir': ds_paths.ldif_dir,
                  'lock_dir': ds_paths.lock_dir,
                  'log_dir': ds_paths.log_dir,
@@ -731,8 +755,24 @@ class SetupDs(object):
             for line in template_dse.readlines():
                 dse += line.replace('%', '{', 1).replace('%', '}', 1)
 
+        # Check if we are in a container, if so don't use /dev/shm for the db home dir
+        # as containers typically don't allocate enough space for dev/shm and we don't
+        # want to unexpectedly break the server after an upgrade
+        #
+        # If we know we are are in a container, we don't need to re-detect on systemd.
+        # It actually turns out if you add systemd-detect-virt, that pulls in system
+        # which subsequently breaks containers starting as instance.start then believes
+        # it COULD check the ds status. The times we need to check for systemd are mainly
+        # in other environments that use systemd natively in their containers.
+        container_result = 1
+        if not self.containerised:
+            container_result = subprocess.run(["systemd-detect-virt", "-c"], stdout=subprocess.PIPE)
+        if self.containerised or container_result.returncode == 0:
+            # In a container, set the db_home_dir to the db path
+            self.log.debug("Container detected setting db home directory to db directory.")
+            slapd['db_home_dir'] = slapd['db_dir']
+
         with open(os.path.join(slapd['config_dir'], 'dse.ldif'), 'w') as file_dse:
-            ldapi_path = os.path.join(slapd['local_state_dir'], "run/slapd-%s.socket" % slapd['instance_name'])
             dse_fmt = dse.format(
                 schema_dir=slapd['schema_dir'],
                 lock_dir=slapd['lock_dir'],
@@ -755,8 +795,9 @@ class SetupDs(object):
                 config_dir=slapd['config_dir'],
                 db_dir=slapd['db_dir'],
                 db_home_dir=slapd['db_home_dir'],
+                db_lib=slapd['db_lib'],
                 ldapi_enabled="on",
-                ldapi=ldapi_path,
+                ldapi=slapd['ldapi'],
                 ldapi_autobind="on",
             )
             file_dse.write(dse_fmt)
@@ -764,7 +805,7 @@ class SetupDs(object):
         self.log.info("Create file system structures ...")
         # Create all the needed paths
         # we should only need to make bak_dir, cert_dir, config_dir, db_dir, ldif_dir, lock_dir, log_dir, run_dir?
-        for path in ('backup_dir', 'cert_dir', 'db_dir', 'ldif_dir', 'lock_dir', 'log_dir', 'run_dir'):
+        for path in ('backup_dir', 'cert_dir', 'db_dir', 'db_home_dir', 'ldif_dir', 'lock_dir', 'log_dir', 'run_dir'):
             self.log.debug("ACTION: creating %s", slapd[path])
             try:
                 os.umask(0o007)  # For parent dirs that get created -> sets 770 for perms
@@ -778,9 +819,16 @@ class SetupDs(object):
         os.chown(parentdir, slapd['user_uid'], slapd['group_gid'])
 
         ### Warning! We need to down the directory under db too for .restore to work.
-        # See dblayer.c for more!
-        db_parent = os.path.join(slapd['db_dir'], '..')
-        os.chown(db_parent, slapd['user_uid'], slapd['group_gid'])
+        # During a restore, the db dir is deleted and recreated, which is why we need
+        # to own it for a restore.
+        #
+        # However, in a container, we can't always guarantee this due to how the volumes
+        # work and are mounted. Specifically, if we have an anonymous volume we will
+        # NEVER be able to own it, but in a true deployment it is reasonable to expect
+        # we DO own it. Thus why we skip it in this specific context
+        if not self.containerised:
+            db_parent = os.path.join(slapd['db_dir'], '..')
+            os.chown(db_parent, slapd['user_uid'], slapd['group_gid'])
 
         # Copy correct data to the paths.
         # Copy in the schema
@@ -838,7 +886,7 @@ class SetupDs(object):
 
         # Should I move this import? I think this prevents some recursion
         from lib389 import DirSrv
-        ds_instance = DirSrv(self.verbose)
+        ds_instance = DirSrv(self.verbose, containerised=self.containerised)
         if self.containerised:
             ds_instance.systemd_override = general['systemd']
 
@@ -858,7 +906,7 @@ class SetupDs(object):
             SER_ROOT_PW: self._raw_secure_password,
             SER_DEPLOYED_DIR: slapd['prefix'],
             SER_LDAPI_ENABLED: 'on',
-            SER_LDAPI_SOCKET: ldapi_path,
+            SER_LDAPI_SOCKET: slapd['ldapi'],
             SER_LDAPI_AUTOBIND: 'on'
         }
 
@@ -898,18 +946,32 @@ class SetupDs(object):
                 selinux_label_port(slapd['secure_port'])
 
         # Do selinux fixups
-        if general['selinux']:
+        if general['selinux'] and selinux_present():
             self.log.info("Perform SELinux labeling ...")
-            selinux_paths = ('backup_dir', 'cert_dir', 'config_dir', 'db_dir',
-                             'ldif_dir', 'lock_dir', 'log_dir',
-                             'run_dir', 'schema_dir', 'tmp_dir')
-            for path in selinux_paths:
-                selinux_restorecon(slapd[path])
+            # Since there may be some custom path, we must explicitly set the labels
+            selinux_labels = {
+                                'backup_dir': 'dirsrv_var_lib_t',
+                                'cert_dir': 'dirsrv_config_t',
+                                'config_dir': 'dirsrv_config_t',
+                                'db_dir': 'dirsrv_var_lib_t',
+                                'ldif_dir': 'dirsrv_var_lib_t',
+                                'lock_dir': 'dirsrv_var_lock_t',
+                                'log_dir': 'dirsrv_var_log_t',
+                                'db_home_dir': 'dirsrv_tmpfs_t',
+                                'run_dir': 'dirsrv_var_run_t',
+                                'schema_dir': 'dirsrv_config_t',
+                                'tmp_dir': 'tmp_t',
+            }
+            for k, label in selinux_labels.items():
+                selinux_label_file(resolve_selinux_path(slapd[k]), label)
 
             selinux_label_port(slapd['port'])
 
         # Start the server
         # Make changes using the temp root
+        self.log.debug(f"asan_enabled={ds_instance.has_asan()}")
+        self.log.debug(f"libfaketime installed ={'libfaketime' in sys.modules}")
+        assert_c(not ds_instance.has_asan() or 'libfaketime' not in sys.modules, "libfaketime python module is incompatible with ASAN build.")
         ds_instance.start(timeout=60)
         ds_instance.open()
 
@@ -951,6 +1013,15 @@ class SetupDs(object):
             self.log.info(f"Create database backend: {backend['nsslapd-suffix']} ...")
             is_sample_entries_in_props = "sample_entries" in backend
             create_suffix_entry_in_props = backend.pop('create_suffix_entry', False)
+            repl_enabled = backend.pop(BACKEND_REPL_ENABLED, False)
+            rid = backend.pop(BACKEND_REPL_ID, False)
+            role = backend.pop(BACKEND_REPL_ROLE, False)
+            binddn = backend.pop(BACKEND_REPL_BINDDN, False)
+            bindpw = backend.pop(BACKEND_REPL_BINDPW, False)
+            bindgrp = backend.pop(BACKEND_REPL_BINDGROUP, False)
+            cl_maxage = backend.pop(BACKEND_REPL_CL_MAX_AGE, False)
+            cl_maxentries = backend.pop(BACKEND_REPL_CL_MAX_ENTRIES, False)
+
             ds_instance.backends.create(properties=backend)
             if not is_sample_entries_in_props and create_suffix_entry_in_props:
                 # Set basic ACIs
@@ -979,6 +1050,74 @@ class SetupDs(object):
                     # Unsupported rdn
                     raise ValueError("Suffix RDN '{}' in '{}' is not supported.  Supported RDN's are: 'c', 'cn', 'dc', 'o', and 'ou'".format(suffix_rdn_attr, backend['nsslapd-suffix']))
 
+            if repl_enabled:
+                # Okay enable replication....
+                self.log.info(f"Enable replication for: {backend['nsslapd-suffix']} ...")
+                repl_root = backend['nsslapd-suffix']
+                if role == "supplier":
+                    repl_type = '3'
+                    repl_flag = '1'
+                elif role == "hub":
+                    repl_type = '2'
+                    repl_flag = '1'
+                elif role == "consumer":
+                    repl_type = '2'
+                    repl_flag = '0'
+                else:
+                    # error - unknown type
+                    raise ValueError("Unknown replication role ({}), you must use \"supplier\", \"hub\", or \"consumer\"".format(role))
+
+                # Start the propeties and update them as needed
+                repl_properties = {
+                    'cn': 'replica',
+                    'nsDS5ReplicaRoot': repl_root,
+                    'nsDS5Flags': repl_flag,
+                    'nsDS5ReplicaType': repl_type,
+                    'nsDS5ReplicaId': '65535'
+                    }
+
+                # Validate supplier settings
+                if role == "supplier":
+                    try:
+                        rid_num = int(rid)
+                        if rid_num < 1 or rid_num > 65534:
+                            raise ValueError
+                        repl_properties['nsDS5ReplicaId'] = rid
+                    except ValueError:
+                        raise ValueError("replica_id expects a number between 1 and 65534")
+
+                    # rid is good add it to the props
+                    repl_properties['nsDS5ReplicaId'] = rid
+
+                # Bind DN & Group
+                if binddn:
+                    repl_properties['nsDS5ReplicaBindDN'] = binddn
+                if bindgrp:
+                    repl_properties['nsDS5ReplicaBindDNGroup'] = bindgrp
+
+                # Enable replication
+                replicas = Replicas(ds_instance)
+                replicas.create(properties=repl_properties)
+
+                # Create replication manager if password was provided
+                if binddn is not None and bindpw:
+                    rdn = binddn.split(",", 1)[0]
+                    rdn_attr, rdn_val = rdn.split("=", 1)
+                    manager = BootstrapReplicationManager(ds_instance, dn=binddn, rdn_attr=rdn_attr)
+                    manager.create(properties={
+                        'cn': rdn_val,
+                        'uid': rdn_val,
+                        'userPassword': bindpw
+                    })
+
+                # Changelog settings
+                if role != "consumer":
+                    cl = Changelog(ds_instance, repl_root)
+                    cl.set_max_age(cl_maxage)
+                    if cl_maxentries != "-1":
+                        cl.set_max_entries(cl_maxentries)
+
+
         # Create all required sasl maps: if we have a single backend ...
         # our default maps are really really bad, and we should feel bad.
         # they basically only work with a single backend, and they'll break
